from typing import Optional

from ..expressions import (
    AddExpression,
    ConstantExpression,
    DivideExpression,
    EqualExpression,
    MathExpression,
    MultiplyExpression,
    SubtractExpression,
)
from ..rule import BaseRule, ExpressionChangeRule
from ..tree import LEFT, RIGHT
from ..types import Literal
from ..util import get_term_ex, unlink

_TYPE_ADDITION = "TYPE_ADDITION"
_TYPE_CONST_OF_MULTIPLY = "TYPE_CONST_OF_MULTIPLY"


class BalancedMoveRule(BaseRule):
    r"""Balanced rewrite rule moves nodes from one side of an equation
    to the other by performing the same operation on both sides.

    Addition: `a + 2 = 3` -> `a + 2 - 2 = 3 - 2`
    Multiplication: `3a = 3` -> `3a / 3 = 3 / 3`
    """

    @property
    def name(self) -> str:
        return "Balanced Move"

    @property
    def code(self) -> str:
        return "BM"

    def has_add_siblings(self, node: MathExpression) -> bool:
        root = node.get_root()
        root_side = node.get_root_side()
        node_subtree = root.left if root_side == LEFT else root.right
        assert node_subtree is not None
        if len(node_subtree.find_type(AddExpression)) > 0:
            return True
        return False

    def get_type(self, node: MathExpression) -> Optional[str]:
        """Determine the configuration of the tree for this transformation.

        Supports the following configurations:
         - Addition is a term connected by an addition to the side of an equation
           or inequality. It generates two subtractions to move from one side to the
           other.
         - Multiply is a coefficient of a term that must be divided on both sides of
           the equation or inequality.
        """
        root = node.get_root()
        if not isinstance(root, EqualExpression) or isinstance(
            node.parent, EqualExpression
        ):
            return None

        if isinstance(node.parent, MultiplyExpression) and isinstance(
            node, ConstantExpression
        ):
            # NOTE: Don't allow divisions or multiplications if there are additions
            #       remaining on the same side of the equation
            if self.has_add_siblings(node):
                return None

            return _TYPE_CONST_OF_MULTIPLY

        if isinstance(node.parent, AddExpression):
            if isinstance(node, ConstantExpression) or get_term_ex(node) is not None:
                return _TYPE_ADDITION

        return None

    def can_apply_to(self, node: MathExpression) -> bool:
        change_type = self.get_type(node)
        return True if change_type is not None else False

    def apply_to(self, node: MathExpression) -> ExpressionChangeRule:
        change = super().apply_to(node)
        change_type = self.get_type(node)
        node = node.clone_from_root()
        root = node.get_root()
        left_new = None
        right_new = None
        if change_type == _TYPE_CONST_OF_MULTIPLY:
            left_new = DivideExpression(root.left, node.clone())
            right_new = DivideExpression(root.right, node.clone())
        elif change_type == _TYPE_ADDITION:
            assert isinstance(node.parent, AddExpression), "expected add parent"
            assert (
                node.parent.parent is not None
            ), "expected node parent to have a parent"
            # If the parent is an Add, this node is a left/right child of an add...
            # You can replace the parent Add with whatever the other sibling is to
            # remove this node.
            save_sibling = unlink(node.get_sibling())
            assert save_sibling is not None
            node_clone = node.clone()
            grandparent = node.parent.parent
            grandparent_side = grandparent.get_side(node.parent)
            update_side: Literal["left", "right"] = (
                LEFT if node.get_root_side() == RIGHT else RIGHT
            )
            grandparent.set_side(save_sibling, grandparent_side)
            new_sub = SubtractExpression(
                root.left if update_side == LEFT else root.right, node_clone
            )
            new_sub.set_changed()
            root.set_side(new_sub, update_side)
            change.done(root)
            return change
        assert left_new is not None, "invalid left node"
        assert right_new is not None, "invalid right node"
        left_new.set_changed()
        right_new.set_changed()
        root.set_left(left_new)
        root.set_right(right_new)
        change.done(root)
        return change



from typing import Optional

from ..expressions import (
    ConstantExpression,
    DivideExpression,
    MathExpression,
    MultiplyExpression,
    NegateExpression,
)
from ..rule import BaseRule, ExpressionChangeRule

_OP_DIVISION_EXPRESSION = "division-expression"
_OP_DIVISION_NEGATIVE_DENOMINATOR = "division-negative-denominator"


class MultiplicativeInverseRule(BaseRule):
    """Convert division operations to multiplication by the reciprocal."""

    @property
    def name(self) -> str:
        return "Multiplicative Inverse"

    @property
    def code(self) -> str:
        return "MI"

    def get_type(self, node: MathExpression) -> Optional[str]:
        """Determine the configuration of the tree for this transformation.

        Support different types of tree configurations based on the division operation:
        - DivisionExpression restated as multiplication by reciprocal
        - DivisionNegativeDenominator is a division by a negative term
        """
        is_division = isinstance(node, DivideExpression)
        if not is_division:
            return None

        # Division where the denominator is negative (e.g., (2 + 3z) / -z)
        if isinstance(node.right, NegateExpression):
            return _OP_DIVISION_NEGATIVE_DENOMINATOR

        # If none of the above, it's a general division expression
        return _OP_DIVISION_EXPRESSION

    def can_apply_to(self, node: MathExpression) -> bool:
        tree_type = self.get_type(node)
        return tree_type is not None

    def apply_to(self, node: MathExpression) -> ExpressionChangeRule:
        change = super().apply_to(node)
        tree_type = self.get_type(node)
        assert tree_type is not None, "call can_apply_to before applying a rule"
        change.save_parent()  # connect result to node.parent

        assert node.left is not None, "Division must have a left child"
        assert node.right is not None, "Division must have a right child"

        # For negative denominator, negate the numerator and use the positive reciprocal
        if tree_type == _OP_DIVISION_NEGATIVE_DENOMINATOR:
            assert isinstance(
                node.right, NegateExpression
            ), "Right child must be a NegateExpression"
            child = node.right.get_child()
            assert child is not None, "NegateExpression must have a child"
            result = MultiplyExpression(
                node.left.clone(),
                DivideExpression(ConstantExpression(-1), child.clone()),
            )
        # Multiply the numerator by the reciprocal of the denominator
        else:
            result = MultiplyExpression(
                node.left.clone(),
                DivideExpression(ConstantExpression(1), node.right.clone()),
            )

        result.set_changed()  # mark this node as changed for visualization
        return change.done(result)



from typing import List

import pytest

from mathy_core import TOKEN_TYPES, Token, Tokenizer


def test_tokenizer_tokenize() -> None:
    text = "4x + 2x^3 * 7!"
    tokenizer = Tokenizer()
    tokens: List[Token] = tokenizer.tokenize(text)

    for token in tokens:
        print(token)
        assert token.type <= TOKEN_TYPES.EOF
        assert token.value is not None


def test_tokenizer_errors() -> None:
    text = "4x + 2x^3 * 7\\"
    tokenizer = Tokenizer()

    with pytest.raises(ValueError):
        tokenizer.tokenize(text)


def test_tokenizer_manual_verification() -> None:
    """Simplest conceptual example verifying the tokenizer maps to
    expectations."""

    manual_tokens: List[Token] = [
        Token("4", TOKEN_TYPES.Constant),
        Token("x", TOKEN_TYPES.Variable),
        Token("+", TOKEN_TYPES.Plus),
        Token("2", TOKEN_TYPES.Constant),
        Token("", TOKEN_TYPES.EOF),
    ]
    auto_tokens: List[Token] = Tokenizer().tokenize("4x + 2")

    for i, token in enumerate(manual_tokens):
        assert auto_tokens[i].value == token.value
        assert auto_tokens[i].type == token.type


def test_tokenizer_identify_functions() -> None:
    """The tokenizer can deal with known function expressions"""
    tokens: List[Token] = Tokenizer().tokenize("sgn(-3)")
    assert tokens[0].type == TOKEN_TYPES.Function
    assert tokens[0].value == "sgn"


def test_tokenizer_ignore_padding() -> None:
    """When specified, the tokenizer will return tokens for padding characters
    to allow exact reproduction of the input string from its tokens."""
    text = "4x + 2y^7 - 6"
    # exclude_padding is True by default
    no_padding: List[Token] = Tokenizer().tokenize(text)
    # When including padding, spaces are preserved
    padding: List[Token] = Tokenizer(exclude_padding=False).tokenize(text)
    assert len(padding) == len(no_padding) + 4



def test_cas_overview_evaluate_expression_variables():
    from ...docs.snippets.cas.overview import evaluate_expression_variables  # noqa


def test_cas_overview_rules_factor_out():
    from ...docs.snippets.cas.overview import rules_factor_out  # noqa


def test_cas_overview_evaluate_expression():
    from ...docs.snippets.cas.overview import evaluate_expression  # noqa



from mathy_core.parser import ExpressionParser
from mathy_core.util import (
    TermEx,
    get_sub_terms,
    get_term_ex,
    get_terms,
    has_like_terms,
    is_preferred_term_form,
    terms_are_like,
)


def test_util_get_sub_terms():
    expectations = [
        ["-f", 1],
        ["70656 * (x^2 * z^6)", 2],
        ["4x^2 * z^6 * y", 3],
        ["2x^2", 1],
        ["x^2", 1],
        ["2", 1],
    ]
    invalid_expectations = [
        # can't have more than one term
        ["4 + 4", False]
    ]
    parser = ExpressionParser()
    for text, output in expectations + invalid_expectations:
        exp = parser.parse(text)
        sub_terms = get_sub_terms(exp)
        if output is False:
            assert text == text and sub_terms == output
        else:
            assert isinstance(sub_terms, list)
            assert text == text and len(sub_terms) == output


def test_util_get_term_ex():
    examples = [
        ["-y", TermEx(-1, "y", None)],
        ["-x^3", TermEx(-1, "x", 3)],
        ["-2x^3", TermEx(-2, "x", 3)],
        ["4x^2", TermEx(4, "x", 2)],
        ["4x", TermEx(4, "x", None)],
        ["x", TermEx(None, "x", None)],
        # TODO: non-natural term forms? If this is supported we can drop the other
        #       get_term impl maybe?
        # ["x * 2", TermEx(2, "x", None)],
    ]
    parser = ExpressionParser()
    for input, expected in examples:
        expr = parser.parse(input)
        assert input == input and get_term_ex(expr) == expected


def test_util_is_preferred_term_form():
    examples = [
        ["b * (44b^2)", False],
        ["z * (1274z^2)", False],
        ["4x * z", True],
        ["z * 4x", True],
        ["2x * x", False],
        ["29y", True],
        ["z", True],
        ["z * 10", False],
        ["4x^2", True],
    ]
    parser = ExpressionParser()
    for input, expected in examples:
        expr = parser.parse(input)
        assert input == input and is_preferred_term_form(expr) == expected


def test_util_has_like_terms():
    examples = [
        ["14 + 6y + 7x + x * (3y)", False],
        ["b * (44b^2)", False],
        ["z * (1274z^2)", False],
        ["100y * x + 2", False],
    ]
    parser = ExpressionParser()
    for input, expected in examples:
        expr = parser.parse(input)
        assert input == input and has_like_terms(expr) == expected


def test_util_terms_are_like():
    parser = ExpressionParser()
    expr = parser.parse("10 + (7x + 6x)")
    terms = get_terms(expr)
    assert len(terms) == 3
    assert not terms_are_like(terms[0], terms[1])
    assert terms_are_like(terms[1], terms[2])

    expr = parser.parse("10 + 7x + 6")
    terms = get_terms(expr)
    assert len(terms) == 3
    assert not terms_are_like(terms[0], terms[1])
    assert terms_are_like(terms[0], terms[2])

    expr = parser.parse("6x + 6 * 5")
    terms = get_terms(expr)
    assert len(terms) == 2
    assert not terms_are_like(terms[0], terms[1])

    expr = parser.parse("360y^1")
    terms = get_terms(expr)
    assert len(terms) == 1

    expr = parser.parse("4z")
    terms = get_terms(expr)
    assert len(terms) == 1



from .associative_swap import AssociativeSwapRule  # noqa
from .balanced_move import BalancedMoveRule  # noqa
from .commutative_swap import CommutativeSwapRule  # noqa
from .constants_simplify import ConstantsSimplifyRule  # noqa
from .distributive_factor_out import DistributiveFactorOutRule  # noqa
from .distributive_multiply_across import DistributiveMultiplyRule  # noqa
from .multiplicative_inverse import MultiplicativeInverseRule  # noqa
from .restate_subtraction import RestateSubtractionRule  # noqa
from .variable_multiply import VariableMultiplyRule  # noqa

__all__ = (
    "AssociativeSwapRule",
    "BalancedMoveRule",
    "CommutativeSwapRule",
    "ConstantsSimplifyRule",
    "DistributiveFactorOutRule",
    "DistributiveMultiplyRule",
    "MultiplicativeInverseRule",
    "RestateSubtractionRule",
    "VariableMultiplyRule",
)



from typing import Callable, Dict, List, Optional, Type

from .expressions import FunctionExpression, SgnExpression
from .types import NumberType


# Define the known types of tokens for the Tokenizer.
class TOKEN_TYPES:
    Constant: int = 1 << 0
    Variable: int = 1 << 1
    Plus: int = 1 << 2
    Minus: int = 1 << 3
    Multiply: int = 1 << 4
    Divide: int = 1 << 5
    Exponent: int = 1 << 6
    Factorial: int = 1 << 7
    OpenParen: int = 1 << 8
    CloseParen: int = 1 << 9
    Function: int = 1 << 10
    Equal: int = 1 << 11
    Pad: int = 1 << 12
    EOF: int = 1 << 13
    Invalid: int = 1 << 14


class Token:
    value: str
    type: int

    def __init__(self, value: str, type: int):
        self.value = value
        self.type = type

    def __str__(self) -> str:
        return f'(type={self.type}, value="{self.value}")'


class TokenContext:
    tokens: List[Token]
    index: int
    buffer: str
    chunk: str

    def __init__(
        self,
        *,
        tokens: Optional[List[Token]] = None,
        index: int = 0,
        buffer: str = "",
        chunk: str = "",
    ):
        self.tokens = tokens if tokens is not None else []
        self.index = index
        self.buffer = buffer
        self.chunk = chunk


class Tokenizer:
    """The Tokenizer produces a list of tokens from an input string."""

    exclude_padding: bool
    functions: Dict[str, Type[FunctionExpression]]

    def __init__(self, exclude_padding: bool = True):
        self.exclude_padding = exclude_padding
        self.functions = {"sgn": SgnExpression}

    def is_alpha(self, c: str) -> bool:
        """Is this character a letter"""
        return ("a" <= c and c <= "z") or ("A" <= c and c <= "Z")

    def is_number(self, c: str) -> bool:
        """Is this character a number"""
        return "." == c or ("0" <= c and c <= "9")

    def eat_token(self, context: TokenContext, typeFn: Callable[[str], bool]) -> str:
        """Eat all of the tokens of a given type from the front of the stream
        until a different type is hit, and return the text."""
        res = ""
        for ch in list(context.chunk):
            if not typeFn(ch):
                return res
            res = res + str(ch)

        return res

    def tokenize(self, buffer: str) -> List[Token]:
        """Return an array of `Token`s from a given string input.
        This throws an exception if an unknown token type is found in the input."""
        context = TokenContext(buffer=buffer, chunk=str(buffer))
        while context.chunk and (
            self.identify_constants(context)
            or self.identify_alphas(context)
            or self.identify_operators(context)
        ):
            context.chunk = context.buffer[context.index :]  # noqa

        context.tokens.append(Token("", TOKEN_TYPES.EOF))
        return context.tokens

    def identify_operators(self, context: TokenContext) -> bool:
        """Identify and tokenize operators."""
        ch = context.chunk[0]
        if ch == " " or ch == "\t" or ch == "\r" or ch == "\n":
            # NOTE: originally introduced this to include padding for token prediction
            if self.exclude_padding is False:
                context.tokens.append(Token(ch, TOKEN_TYPES.Pad))
        elif ch == "+":
            context.tokens.append(Token("+", TOKEN_TYPES.Plus))
        elif ch == "-" or ch == "–":
            context.tokens.append(Token("-", TOKEN_TYPES.Minus))
        elif ch == "*":
            context.tokens.append(Token("*", TOKEN_TYPES.Multiply))
        elif ch == "/":
            context.tokens.append(Token("/", TOKEN_TYPES.Divide))
        elif ch == "^":
            context.tokens.append(Token("^", TOKEN_TYPES.Exponent))
        elif ch == "!":
            context.tokens.append(Token("!", TOKEN_TYPES.Factorial))
        elif ch == "(" or ch == "[":
            context.tokens.append(Token("(", TOKEN_TYPES.OpenParen))
        elif ch == ")" or ch == "]":
            context.tokens.append(Token(")", TOKEN_TYPES.CloseParen))
        elif ch == "=":
            context.tokens.append(Token("=", TOKEN_TYPES.Equal))
        else:
            raise ValueError(f'Invalid token "{ch}" in expression: {context.buffer}')
        context.index = context.index + 1
        return True

    def identify_alphas(self, context: TokenContext) -> int:
        """Identify and tokenize functions and variables."""
        if not self.is_alpha(context.chunk[0]):
            return False

        variable = self.eat_token(context, self.is_alpha)
        if variable in self.functions:
            context.tokens.append(Token(variable, TOKEN_TYPES.Function))
        else:
            # Each letter is its own variable
            for c in variable:
                context.tokens.append(Token(c, TOKEN_TYPES.Variable))

        context.index += len(variable)
        return len(variable)

    def identify_constants(self, context: TokenContext) -> int:
        """Identify and tokenize a constant number."""
        if not self.is_number(context.chunk[0]):
            return 0

        val = self.eat_token(context, self.is_number)
        context.tokens.append(Token(val, TOKEN_TYPES.Constant))
        context.index += len(val)
        return len(val)


def coerce_to_number(value: str) -> NumberType:
    return float(value) if "e" in value or "." in value else int(value)


__all__ = ("TOKEN_TYPES", "Token", "TokenContext", "Tokenizer", "coerce_to_number")
